/*====================================================================================================*
* PINEWOOD DERBY - PWD 2-Lane Track Timer                                    Version 1b - 06 Feb 2018 *
* Created by: Bill van Leeuwen (email: Billv923@outlook.com) for the Arduino UNO microcontroller      *
*                                                                                                     *
* This program is designed to work in conjunction with the PWD Track Timer display software (created  *
* by the above author) which runs on a PC attached to the Arduino via the USB serial port.            *
*                                                                                                     *
* This program measures the start-to-finish time of 2 lanes. Race status and finish times are sent to * 
* the serial port to interface with the PWD Track Timer display software running on the PC. Time is   *
* measured from activation of the start switch to activation of the finish line sensor of each lane.  *
* Each car has 10 seconds to finish the race. If a car doesn't cross the finish line within 10        *
* seconds, it is assumed that the car had a problem or the lane was not used. In either case, the     *
* value 9.9999 is assigned to the lane time as an error indication for use by the PWD Track Timer     *
* display software running on the PC.                                                                 *
*                                                                                                     *
* This design uses infrared sensors (one for each lane) imbedded in the track at the finish line to   *
* detect when a car crosses the finish line. Incandescent lamps mounted above the finish line provide *
* the illumination needed by the sensors. A switch connected to the start gate is used to start the   *
* timer. Lane times and track status codes are sent via the USB port to the PWD Track Timer display   *
* software running on the PC.                                                                         *
*                                                                                                     *
* Refer to the Schematic Diagram, parts list and design notes that accompany this software for wiring * 
* of the infrared lane sensors, race start switch, and reset switch.                                  *
* Feel free to contact me at Billv923@outlook.com if any questions.                                   *
*====================================================================================================*/

//Initialization
#define Num_Lanes    2         //Number of lanes
#define NotReady     0         //Race State = Not Ready
#define Ready        1         //Race State = Ready
#define Racing       2         //Race State = Racing
#define Finished     3         //Race State = Finished
#define Msg_Reset   'R'        //Reset character received from PC via Serial port

//Pin Assignments
#define fs1 2                  //Finish Sensor Lane 1 - Digital Input Pin 2
#define fs2 3                  //Finish Sensor Lane 2 - Digital Input Pin 3
#define gate_sw 8              //Gate Switch (Start)  - Digital Input Pin 8
#define reset_sw 9             //Manual Reset         - Digital Input Pin 9
#define LED 13                 //IO Pin 13 LED        - Digital Pin 13


//Variable Initialization
int i;                         //Working variable
int Lane_Place[Num_Lanes];     //Lane finish place array
boolean sens_ready;            //False if any lane sensor is obstructed
boolean track_ready;           //False if any lane sensor is obstructed or Gate Switch Open
int lane;                      //Lane number
int finish_order;              //Finish order
int serial_data;               //Serial data
int racestate;                 //0=NotReady, 1=Ready, 2=Racing, 3=Finished
int fincount;                  //Tallies cars crossing finish line to determine race over
unsigned long Start_Time;      //Race Start (Gate Release)Time (microseconds)
unsigned long Exp_Time;        //Expiration Time (milliseconds)
unsigned long Clk_Time;        //CPU Clock Time (microseconds)
float Time;                    //Working variable - Calculated Lane Time (seconds)
float Lane_Time[Num_Lanes];    //Calculated lane time data array (seconds)
char SMsg0[4] = "NRD";         //Serial message - Not Ready
char SMsg1[4] = "RDY";         //Serial message - Ready
char SMsg2[4] = "RAC";         //Serial message - Racing
char SMsg3[4] = "FIN";         //Serial message - Finished
char SMsg4[4] = "GSW";         //Serial message - Gate Switch not ready
char SMsg5[4] = "TRK";         //Serial message - Track Sensor(s) blocked
String bad_sens = "";          //Message string populated with obstructed lane sensors

//======================================= SETUP ROUTINE =======================================
void setup() {  
//Hardware Setup
  //Initialize Finish Sensor, Gate Switch and Reset Switch pins as inputs & pin 13 as an output
  pinMode(fs1,INPUT_PULLUP);          //Note: Sensor inputs fs1 thru fs2 transition from LOW to
  pinMode(fs2,INPUT_PULLUP);          //      HIGH when light beam is interupted
  pinMode(gate_sw,INPUT_PULLUP);      //Gate_sw is normally LOW & transitions to HIGH at race start
  pinMode(reset_sw,INPUT_PULLUP);     //Reset_sw is normally HIGH & transitions to LOW when pressed
  pinMode(LED, OUTPUT);               //set so we can turn off the pin 13 LED
  
//Software Setup
  Serial.begin(9600);                 //Enable serial comms to PC at 9600 baud
  racestate = NotReady;               //Set initial race state to not ready
  digitalWrite(LED,LOW);              //Turn off Pin 13 LED just for the heck of it
}

//======================================== MAIN LOOP ============================================
void loop() {
  //READ & PROCESS INCOMING SERIAL DATA
  serial_data = get_serial_data();                //Get any received serial data
  if (serial_data == int(Msg_Reset)) {            //Check for Reset request from PC
    Check_Track();                                //If so, call Check Track subroutine
    if (track_ready){                             //and only do a reset if Track is ready
      initialize();                               //Call initialize function & reset everything
      racestate = Ready;                          //Set racestate to Ready
      Serial.println(SMsg1);                      //Send Ready state message to PC
    } 
    else {                                        //Do this if track is not ready
      racestate = NotReady;                       //Set racestate to Not Ready
      Serial.println(SMsg0);                      //Send Not Ready state message to PC
    }
  }
 
  //CHECK & PROCESS RACE STATES
  if (digitalRead(reset_sw)==0){                  //Manual timer Reset button pressed?
    Check_Track();                                //If so, call Check Track subroutine
    if (track_ready){                             //and only do a reset if Track is ready
      initialize();                               //Call initialize function & reset everything
      racestate = Ready;                          //Set racestate to Ready
      Serial.println(SMsg1);                      //Send Ready state message to PC
    } 
    else {                                        //Do this if track is not ready
      racestate = NotReady;                       //Set racestate to NotReady
      Serial.println(SMsg0);                      //Send Not Ready state message to PC
    }
    delay(500);                                   //Insert delay to debounce reset switch
  }
  // Transition from Ready to Racing when the Gate switch is tripped
  if (racestate == Ready && digitalRead(gate_sw)==1) {
    Start_Time = micros();                        //Get start time in microseconds
    Exp_Time = Start_Time + 10000000;             //Set Expiration = Start_Time + 10 seconds (in microseconds)
    racestate = Racing;                           //Set racestate to Racing
    Serial.println(SMsg2);                        //Send Racing state message to PC
    delay(200);                                   //Wait a while before we start finish check loop
    
    //Race underway. Check/Wait for everyone to finish
    /*====================================================================================================*
    * NOTE!!!!: The 'while' loop below has been optimized to repeatedly check lane sensors as fast as     *
    * possible by accessing the digital port input register directly via the command 'bitRead(PIND,fsx)'  *
    * instead of the command 'digitalRead()'. This command is specifically tailored to Arduino boards     *
    * using the ATmega8, ATmega168 or ATmega328p chips. If this code is used with an Arduino clone having *
    * a different processing chip, it may not work. If this is the case, replace each occurence of the    *
    * 'bitRead(PIND,fsx)' command with 'digitalRead(fsx)' where x is the lane number.                     *
    *====================================================================================================*/
    while (fincount<Num_Lanes) {                  //Loop until all cars have crossed finish line
      Clk_Time = micros();                        //Get the CPU clock time in microseconds
      if (Lane_Place[0]==0) {                     //Check & skip if Lane 1 already finished
        if (bitRead(PIND, fs1)) {                 //Is Car over Lane 1 Finish Line?
          Lane_Time[0]=Clk_Time;                  //Yes...save the time
          Lane_Place[0]=++fincount;               //Incr finish count & set finish place for Lane 1
        }
      }
      if (Lane_Place[1]==0) {                     //Check & skip if Lane 2 already finished
        if (bitRead(PIND, fs2)) {                 //Is Car over Lane 2 Finish Line?
          Lane_Time[1]=Clk_Time;                  //Yes...save the time
          Lane_Place[1]=++fincount;               //Incr finish count & set finish place for Lane 2
        }
      }  
      if (Clk_Time>Exp_Time) break;               //Exit 'while' loop if time has expired
    }
    //Race is over or time has expired
    racestate=Finished;                           //Set racestate to Finished
    //Send Race Results To PC as one long string (done to simplify processing at the PC)
    Serial.print("Times:");                       //Precede string with "Times" as an identifier
    for (int n=0; n<Num_Lanes; n++) {             //Loop thru all lanes
      Serial.print(" ");                          //Add a space to separate numbers
      if (Lane_Time[n]>0) {                       //Check to see if lane time is valid
        Time = (float)((Lane_Time[n]-Start_Time)/1000000.0);  //Calc lane time and convert to seconds
        Serial.print(Time,4);                     //Print lane time trimmed to 4 places after decimal
      }
      else {
        Serial.print("9.9999");                   //Print all nines if lane time is invalid
      }
    }
    Serial.println("");                           //Send a cr/lf to indicate end of the string
    delay(100);                                   //Wait a while to allow display SW to process times
    Serial.println(SMsg3);                        //Send Finish state message to PC
    delay(100);                                   //Wait a while to allow display SW to process SMsg3
  }                                               //End Racing "if"
}                                                 //Loop Back

/*================================================================================*
  SUBROUTINES & FUNCTIONS
 *================================================================================*/
//Subroutine: Initialize Racing Variables
void initialize() {  
  for (int n=0; n<Num_Lanes; n++) {           //Reset lane Place & Times
    Lane_Place[n] = 0;
    Lane_Time[n] = 0;
  }
  fincount = 0;                               //Reset finish count to 0
  Serial.flush();                             //Clear the serial input buffer
  return;
}

//Function: Read Serial Data
int get_serial_data() {
  int data = 0;                               //Clear any old data
  if (Serial.available() > 0) {               //Check for new data
    data = Serial.read();                     //and get it if so
  }
  return data;
}  

//Function: Check Track - Sets 'track_ready' to true or false
void Check_Track() {
  track_ready = true;                         //Default track ready status to true
  sens_ready = true;                          //Default sensor ready status to true
  bad_sens = "";                              //Clear bad sensor string
  if (digitalRead(gate_sw)==1) {              //Check if Gate Switch is open
    Serial.println(SMsg4);                    //If so, send Gate Switch not ready message to PC
    track_ready = false;                      //And set track ready status to false
  }
  for (int n=0; n<Num_Lanes; n++) {           //Loop through all lanes and
    if (digitalRead(n+2)==1) {                //Check if a lane sensor (pins 2 & 3) is obstructed
      sens_ready = false;                     //If so, set sensor ready status to false
      bad_sens = bad_sens + ", " + (n+1);     //Build string identifying obstructed lane sensors
    }
  }
  if (sens_ready == false) {                  //Check if a lane sensor was not ready
    Serial.println(SMsg5 + bad_sens);         //If so, send Lane Sensors not ready message to PC
    track_ready = false;                      //And set track ready status to false
  }
  delay(200);                                 //Wait a while to allow display SW to process messages
}


